"""
Interface to support treating videos like arrays using OpenCV
"""

from pathlib import Path
from typing import Any, Literal, Optional, Tuple, Union

import numpy as np
from pydantic_core.core_schema import SerializationInfo

from numpydantic.interface import JsonDict
from numpydantic.interface.interface import Interface

try:
    import cv2
    from cv2 import VideoCapture
except ImportError:  # pragma: no cover
    cv2 = None
    VideoCapture = None

VIDEO_EXTENSIONS = (".mp4", ".avi", ".mov", ".mkv")


class VideoJsonDict(JsonDict):
    """Json-able roundtrip representation of a video file"""

    type: Literal["video"]
    file: str

    def to_array_input(self) -> "VideoProxy":
        """
        Construct a :class:`.VideoProxy`
        """
        return VideoProxy(path=Path(self.file))


class VideoProxy:
    """
    Passthrough proxy class to interact with videos as arrays
    """

    def __init__(
        self, path: Optional[Path] = None, video: Optional[VideoCapture] = None
    ):
        if path is None and video is None:  # pragma: no cover
            raise ValueError(
                "Need to either supply a path or an opened VideoCapture object"
            )

        if path is not None:
            path = Path(path).resolve()
        self.path = path

        self._video = video  # type: Optional[VideoCapture]
        self._n_frames = None  # type: Optional[int]
        self._dtype = None  # type: Optional[np.dtype]
        self._shape = None  # type: Optional[Tuple[int, ...]]
        self._sample_frame = None  # type: Optional[np.ndarray]

    @property
    def video(self) -> VideoCapture:
        """Opened video capture object"""
        if self._video is None:
            if self.path is None:  # pragma: no cover
                raise RuntimeError(
                    "Instantiated with a VideoCapture object that has been closed, "
                    "and it cant be reopened since source path cant be gotten "
                    "from VideoCapture objects"
                )
            if not self.path.exists():
                raise FileNotFoundError(f"Video file {self.path} does not exist!")

            self._video = VideoCapture(str(self.path))
        return self._video

    def close(self) -> None:
        """Close the opened VideoCapture object"""
        if self._video is not None:
            self._video.release()
            self._video = None

    @property
    def sample_frame(self) -> np.ndarray:
        """A stored frame from the video to use when calculating shape and dtype"""
        if self._sample_frame is None:
            current_frame = int(self.video.get(cv2.CAP_PROP_POS_FRAMES))

            self.video.set(cv2.CAP_PROP_POS_FRAMES, max(0, current_frame - 1))
            status, frame = self.video.read()
            if not status:  # pragma: no cover
                raise RuntimeError("Could not read frame from video")
            self.video.set(cv2.CAP_PROP_POS_FRAMES, current_frame)
            self._sample_frame = frame
        return self._sample_frame

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        Shape of video like
        ``(n_frames, height, width, channels)``

        Note that this order flips the order of height and width from typical resolution
        specifications: eg. 1080p video is typically 1920x1080, but here it would be
        1080x1920. This follows opencv's ordering, which matches expectations when
        eg. an image is read and plotted with matplotlib: the first index is the
        position in the 0th dimension - the height, or "y" axis - and the second is the
        width/x.
        """
        if self._shape is None:
            self._shape = (self.n_frames, *self.sample_frame.shape)
        return self._shape

    @property
    def dtype(self) -> np.dtype:
        """Numpy dtype (from ``sample_frame`` )"""
        return self.sample_frame.dtype

    @property
    def n_frames(self) -> int:
        """
        Try to get number of frames using opencv metadata, and manually count if no
        t"""
        if self._n_frames is None:
            n_frames = self.video.get(cv2.CAP_PROP_FRAME_COUNT)
            if n_frames == 0:  # pragma: no cover
                # have to count manually for some containers with bad metadata
                # not testing for now, will wait until we encounter such a
                # video in the wild where this doesn't work.
                current_frame = self.video.get(cv2.CAP_PROP_POS_FRAMES)
                self.video.set(cv2.CAP_PROP_POS_FRAMES, 0)
                n_frames = 0
                while True:
                    status, _ = self.video.read()
                    if not status:
                        break
                    n_frames += 1
                self.video.set(cv2.CAP_PROP_POS_FRAMES, current_frame)
            self._n_frames = int(n_frames)
        return self._n_frames

    def _get_frame(self, frame: int) -> np.ndarray:
        self.video.set(cv2.CAP_PROP_POS_FRAMES, frame)
        status, frame = self.video.read()
        if not status:  # pragma: no cover
            raise ValueError(f"Could not get frame {frame}")
        return frame

    def _complete_slice(self, slice_: slice) -> slice:
        """Get a fully-built slice that can be passed to range"""
        if slice_.step is None:
            slice_ = slice(slice_.start, slice_.stop, 1)
        if slice_.stop is None:
            slice_ = slice(slice_.start, self.n_frames, slice_.step)
        if slice_.start is None:
            slice_ = slice(0, slice_.stop, slice_.step)
        return slice_

    def __array__(self) -> np.ndarray:
        """Whole video as a numpy array"""
        return self[:]

    def __getitem__(self, item: Union[int, slice, tuple]) -> np.ndarray:
        if isinstance(item, int):
            # want a single frame
            return self._get_frame(item)
        elif isinstance(item, slice):
            # slice of frames
            item = self._complete_slice(item)
            frames = [
                self._get_frame(i) for i in range(item.start, item.stop, item.step)
            ]
            return np.stack(frames)
        else:
            # slices are passed as tuples
            # first arg needs to be handled specially
            if isinstance(item[0], int):
                # single frame
                frame = self._get_frame(item[0])
                # syntax doesn't work in 3.9 but would be more explicit...
                # return frame[*item[1:]]
                return frame[item[1:]]

            elif isinstance(item[0], slice):
                frames = []
                # make a new slice since range cant take Nones, filling in missing vals
                fslice = self._complete_slice(item[0])

                for i in range(fslice.start, fslice.stop, fslice.step):
                    frames.append(self._get_frame(i))
                frame = np.stack(frames)
                # syntax doesn't work in 3.9 but would be simpler..
                # return frame[:, *item[1:]]
                # construct a new slice instead
                new_slice = (slice(None, None, None), *item[1:])
                return frame[new_slice]
            else:  # pragma: no cover
                raise ValueError(f"indices must be an int or a slice! got {item}")

    def __setitem__(self, key: Union[int, slice], value: Union[int, float, np.ndarray]):
        raise NotImplementedError("Setting pixel values on videos is not supported!")

    def __getattr__(self, item: str):
        if item == "__name__":
            return "VideoProxy"
        return getattr(self.video, item)

    def __eq__(self, other: "VideoProxy") -> bool:
        """Check if this is a proxy to the same video file"""
        if not isinstance(other, VideoProxy):
            raise TypeError("Can only compare equality of two VideoProxies")
        return self.path == other.path

    def __len__(self) -> int:
        """Number of frames in the video"""
        return self.shape[0]


class VideoInterface(Interface):
    """
    OpenCV interface to treat videos as arrays.
    """

    name = "video"
    input_types = (str, Path, VideoCapture, VideoProxy)
    return_type = VideoProxy
    json_model = VideoJsonDict

    @classmethod
    def enabled(cls) -> bool:
        """Check if opencv-python is available in the environment"""
        return cv2 is not None

    @classmethod
    def check(cls, array: Any) -> bool:
        """
        Check if array is a string or Path with a supported video extension,
        or an opened VideoCapture object
        """
        if (VideoCapture is not None and isinstance(array, VideoCapture)) or isinstance(
            array, VideoProxy
        ):
            return True

        if isinstance(array, dict):
            array = array.get("file", "")

        if isinstance(array, str):
            try:
                array = Path(array)
            except TypeError:  # pragma: no cover
                # fine, just not a video
                return False

        return isinstance(array, Path) and array.suffix.lower() in VIDEO_EXTENSIONS

    def before_validation(self, array: Any) -> VideoProxy:
        """Get a :class:`.VideoProxy` object for this video"""
        if isinstance(array, VideoCapture):
            proxy = VideoProxy(video=array)
        elif isinstance(array, VideoProxy):
            proxy = array
        else:
            proxy = VideoProxy(path=array)
        return proxy

    @classmethod
    def to_json(
        cls, array: VideoProxy, info: SerializationInfo
    ) -> Union[list, VideoJsonDict]:
        """Return a json-representation of a video"""
        if info.round_trip:
            return VideoJsonDict(type=cls.name, file=str(array.path))
        else:
            return np.array(array).tolist()



"""
Base Interface metaclass
"""

import inspect
import warnings
from abc import ABC, abstractmethod
from functools import lru_cache
from importlib.metadata import PackageNotFoundError, version
from operator import attrgetter
from typing import Any, Generic, Optional, Tuple, Type, TypeVar, Union

import numpy as np
from pydantic import BaseModel, SerializationInfo, ValidationError

from numpydantic.exceptions import (
    DtypeError,
    MarkMismatchError,
    NoMatchError,
    ShapeError,
    TooManyMatchesError,
)
from numpydantic.types import DtypeType, NDArrayType, ShapeType
from numpydantic.validation import validate_dtype, validate_shape

T = TypeVar("T", bound=NDArrayType)
U = TypeVar("U", bound="JsonDict")
V = TypeVar("V")  # input type
W = TypeVar("W")  # Any type in handle_input


class InterfaceMark(BaseModel):
    """JSON-able mark to be able to round-trip json dumps"""

    module: str
    cls: str
    name: str
    version: str

    def is_valid(self, cls: Type["Interface"], raise_on_error: bool = False) -> bool:
        """
        Check that a given interface matches the mark.

        Args:
            cls (Type): Interface type to check
            raise_on_error (bool): Raise an ``MarkMismatchError`` when the match
                is incorrect

        Returns:
            bool

        Raises:
            :class:`.MarkMismatchError` if requested by ``raise_on_error``
            for an invalid match
        """
        mark = cls.mark_interface()
        valid = self == mark
        if not valid and raise_on_error:
            raise MarkMismatchError(
                "Mismatch between serialized mark and current interface, "
                f"Serialized: {self}; current: {cls}"
            )
        return valid

    def match_by_name(self) -> Optional[Type["Interface"]]:
        """
        Try to find a matching interface by its name, returning it if found,
        or None if not found.
        """
        for i in Interface.interfaces(sort=False):
            if i.name == self.name:
                return i
        return None


class JsonDict(BaseModel):
    """
    Representation of array when dumped with round_trip == True.

    .. admonition:: Developer's Note

        Any JsonDict that contains an actual array should be named ``value``
        rather than array (or any other name), and nothing but the
        array data should be named ``value`` .

        During JSON serialization, it becomes ambiguous what contains an array
        of data vs. an array of metadata. For the moment we would like to
        reserve the ability to have lists of metadata, so until we rule that out,
        we would like to be able to avoid iterating over every element of an array
        in any context parameter transformation like relativizing/absolutizing paths.
        To avoid that, it's good to agree on a single value name -- ``value`` --
        and avoid using it for anything else.

    """

    type: str

    @abstractmethod
    def to_array_input(self) -> V:
        """
        Convert this roundtrip specifier to the relevant input class
        (one of the ``input_types`` of an interface).
        """

    @classmethod
    def is_valid(cls, val: dict, raise_on_error: bool = False) -> bool:
        """
        Check whether a given dictionary matches this JsonDict specification

        Args:
            val (dict): The dictionary to check for validity
            raise_on_error (bool): If ``True``, raise the validation error
                rather than returning a bool. (default: ``False``)

        Returns:
            bool - true if valid, false if not
        """
        try:
            _ = cls.model_validate(val)
            return True
        except ValidationError as e:
            if raise_on_error:
                raise e
            return False

    @classmethod
    def handle_input(cls: Type[U], value: Union[dict, U, W]) -> Union[V, W]:
        """
        Handle input that is the json serialized roundtrip version
        (from :func:`~pydantic.BaseModel.model_dump` with ``round_trip=True``)
        converting it to the input format with :meth:`.JsonDict.to_array_input`
        or passing it through if not applicable
        """
        if isinstance(value, dict):
            value = cls(**value).to_array_input()
        elif isinstance(value, cls):
            value = value.to_array_input()
        return value

    @staticmethod
    def reshape_input(value: T, shape: tuple[int, ...]) -> T:
        """
        If a `reshape` value is present on the array, and the array shape doesn't match,
        attempt to reshape it.
        """
        if value.shape != shape:
            try:
                value = value.reshape(shape)
            except ValueError:
                warnings.warn(
                    f"Input data has shape {value.shape}, "
                    f"but roundtrip form specifies {shape},"
                    f"and {value.shape} can't be cast to {shape}. "
                    f"Attempting to proceed with validation without reshaping.",
                    stacklevel=1,
                )
        return value


class MarkedJson(BaseModel):
    """
    Model of JSON dumped with an additional interface mark
    with ``model_dump_json({'mark_interface': True})``
    """

    interface: InterfaceMark
    value: Union[list, dict]
    """
    Inner value of the array, we don't validate for JsonDict here, 
    that should be downstream from us for performance reasons 
    """

    @classmethod
    def try_cast(cls, value: Union[V, dict]) -> Union[V, "MarkedJson"]:
        """
        Try to cast to MarkedJson if applicable, otherwise return input
        """
        if isinstance(value, dict) and "interface" in value and "value" in value:
            try:
                value = MarkedJson(**value)
            except ValidationError:
                # fine, just not a MarkedJson dict even if it looks like one
                return value
        return value


class Interface(ABC, Generic[T]):
    """
    Abstract parent class for interfaces to different array formats
    """

    input_types: Tuple[Any, ...]
    return_type: Type[T]
    priority: int = 0

    def __init__(self, shape: ShapeType = Any, dtype: DtypeType = Any) -> None:
        self.shape = shape
        self.dtype = dtype

    def validate(self, array: Any) -> T:
        """
        Validate input, returning final array type

        Calls the methods, in order:

        * array = :meth:`.deserialize` (array)
        * array = :meth:`.before_validation` (array)
        * dtype = :meth:`.get_dtype` (array) - get the dtype from the array,
            override if eg. the dtype is not contained in ``array.dtype``
        * valid = :meth:`.validate_dtype` (dtype) - check that the dtype matches
            the one in the NDArray specification. Override if special
            validation logic is needed for a given format
        * :meth:`.raise_for_dtype` (valid, dtype) - after checking dtype validity,
            raise an exception if it was invalid. Override to implement custom
            exceptions or error conditions, or make validation errors conditional.
        * array = :meth:`.after_validate_dtype` (array) - hook for additional
            validation or array modification mid-validation
        * shape = :meth:`.get_shape` (array) - get the shape from the array,
            override if eg. the shape is not contained in ``array.shape``
        * valid = :meth:`.validate_shape` (shape) - check that the shape matches
            the one in the NDArray specification. Override if special validation
            logic is needed.
        * :meth:`.raise_for_shape` (valid, shape) - after checking shape validity,
            raise an exception if it was invalid. You know the deal bc it's the same
            as raise for dtype.
        * :meth:`.after_validation` - hook after validation for modifying the array
            that is set as the model field value

        Follow the method signatures and return types to override.

        Implementing an interface subclass largely consists of overriding these methods
        as needed.

        Raises:
            If validation fails, rather than eg. returning ``False``, exceptions will
            be raised (to halt the rest of the pydantic validation process).
            When using interfaces outside of pydantic, you must catch both
            :class:`.DtypeError` and :class:`.ShapeError` (both of which are children
            of :class:`.InterfaceError` )
        """
        array = self.deserialize(array)

        array = self.before_validation(array)

        dtype = self.get_dtype(array)
        dtype_valid = self.validate_dtype(dtype)
        self.raise_for_dtype(dtype_valid, dtype)
        array = self.after_validate_dtype(array)

        shape = self.get_shape(array)
        shape_valid = self.validate_shape(shape)
        self.raise_for_shape(shape_valid, shape)

        array = self.after_validation(array)

        return array

    def deserialize(self, array: Any) -> Union[V, Any]:
        """
        If given a JSON serialized version of the array,
        deserialize it first.

        If a roundtrip-serialized :class:`.JsonDict`,
        pass to :meth:`.JsonDict.handle_input`.

        If a roundtrip-serialized :class:`.MarkedJson`,
        unpack mark, check for validity, warn if not,
        and try to continue with validation
        """
        if isinstance(marked_array := MarkedJson.try_cast(array), MarkedJson):
            try:
                marked_array.interface.is_valid(self.__class__, raise_on_error=True)
            except MarkMismatchError as e:
                warnings.warn(
                    str(e) + "\nAttempting to continue validation...", stacklevel=2
                )
            array = marked_array.value

        return self.json_model.handle_input(array)

    def before_validation(self, array: Any) -> NDArrayType:
        """
        Optional step pre-validation that coerces the input into a type that can be
        validated for shape and dtype

        Default method is a no-op
        """
        return array

    def get_dtype(self, array: NDArrayType) -> DtypeType:
        """
        Get the dtype from the input array.
        """
        if hasattr(array.dtype, "type") and array.dtype.type is np.object_:
            return self.get_object_dtype(array)
        else:
            return array.dtype

    def get_object_dtype(self, array: NDArrayType) -> DtypeType:
        """
        When an array contains an object, get the dtype of the object contained
        by the array.

        If this method returns `Any`, the dtype validation passes -
        used for e.g. empty arrays for which the dtype of the array can't be determined
        (since there are no objects).
        """
        try:
            return type(array.ravel()[0])
        except IndexError:
            return Any

    def validate_dtype(self, dtype: DtypeType) -> bool:
        """
        Validate the dtype of the given array, returning
        ``True`` if valid, ``False`` if not.
        """
        return validate_dtype(dtype, self.dtype)

    def raise_for_dtype(self, valid: bool, dtype: DtypeType) -> None:
        """
        After validating, raise an exception if invalid
        Raises:
            :class:`~numpydantic.exceptions.DtypeError`
        """
        if not valid:
            raise DtypeError(f"Invalid dtype! expected {self.dtype}, got {dtype}")

    def after_validate_dtype(self, array: NDArrayType) -> NDArrayType:
        """
        Hook to modify array after validating dtype.
        Default is a no-op.
        """
        return array

    def get_shape(self, array: NDArrayType) -> Tuple[int, ...]:
        """
        Get the shape from the array as a tuple of integers
        """
        return array.shape

    def validate_shape(self, shape: Tuple[int, ...]) -> bool:
        """
        Validate the shape of the given array against the shape
        specifier, returning ``True`` if valid, ``False`` if not.


        """
        if self.shape is Any:
            return True

        return validate_shape(shape, self.shape)

    def raise_for_shape(self, valid: bool, shape: Tuple[int, ...]) -> None:
        """
        Raise a ShapeError if the shape is invalid.

        Raises:
            :class:`~numpydantic.exceptions.ShapeError`
        """
        if not valid:
            raise ShapeError(
                f"Invalid shape! expected shape {self.shape.prepared_args}, "
                f"got shape {shape}"
            )

    def after_validation(self, array: NDArrayType) -> T:
        """
        Optional step post-validation that coerces the intermediate array type into the
        return type

        Default method is a no-op
        """
        return array

    @classmethod
    @abstractmethod
    def check(cls, array: Any) -> bool:
        """
        Method to check whether a given input applies to this interface
        """

    @classmethod
    @abstractmethod
    def enabled(cls) -> bool:
        """
        Check whether this array interface can be used (eg. its dependent packages are
        installed, etc.)
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Short name for this interface
        """

    @property
    @abstractmethod
    def json_model(self) -> JsonDict:
        """
        The :class:`.JsonDict` model used for roundtripping
        JSON serialization
        """

    @classmethod
    @abstractmethod
    def to_json(cls, array: Type[T], info: SerializationInfo) -> Union[list, JsonDict]:
        """
        Convert an array of :attr:`.return_type` to a JSON-compatible format using
        base python types
        """

    @classmethod
    def mark_json(cls, array: Union[list, dict]) -> dict:
        """
        When using ``model_dump_json`` with ``mark_interface: True`` in the ``context``,
        add additional annotations that would allow the serialized array to be
        roundtripped.

        Default is just to add an :class:`.InterfaceMark`

        Examples:

            >>> from pprint import pprint
            >>> pprint(Interface.mark_json([1.0, 2.0]))
            {'interface': {'cls': 'Interface',
                           'module': 'numpydantic.interface.interface',
                           'version': '1.2.2'},
             'value': [1.0, 2.0]}
        """
        return {"interface": cls.mark_interface(), "value": array}

    @classmethod
    def interfaces(
        cls, with_disabled: bool = False, sort: bool = True
    ) -> Tuple[Type["Interface"], ...]:
        """
        Enabled interface subclasses

        Args:
            with_disabled (bool): If ``True`` , get every known interface.
                If ``False`` (default), get only enabled interfaces.
            sort (bool): If ``True`` (default), sort interfaces by priority.
                If ``False`` , sorted by definition order. Used for recursion:
                we only want to sort once at the top level.
        """
        # get recursively
        subclasses = []
        for i in cls.__subclasses__():
            if with_disabled:
                subclasses.append(i)

            if i.enabled():
                subclasses.append(i)

            subclasses.extend(i.interfaces(with_disabled=with_disabled, sort=False))

        if sort:
            subclasses = sorted(
                subclasses,
                key=attrgetter("priority"),
                reverse=True,
            )

        return tuple(subclasses)

    @classmethod
    def return_types(cls) -> Tuple[NDArrayType, ...]:
        """Return types for all enabled interfaces"""
        return tuple([i.return_type for i in cls.interfaces()])

    @classmethod
    def input_types(cls) -> Tuple[Any, ...]:
        """Input types for all enabled interfaces"""
        in_types = []
        for iface in cls.interfaces():
            if isinstance(iface.input_types, (tuple, list)):
                in_types.extend(iface.input_types)
            else:  # pragma: no cover
                in_types.append(iface.input_types)

        return tuple(in_types)

    @classmethod
    def match_mark(cls, array: Any) -> Optional[Type["Interface"]]:
        """
        Match a marked JSON dump of this array to the interface that it indicates.

        First find an interface that matches by name, and then run its
        ``check`` method, because arrays can be dumped with a mark
        but without ``round_trip == True`` (and thus can't necessarily
        use the same interface that they were dumped with)

        Returns:
            Interface if match found, None otherwise
        """
        mark = MarkedJson.try_cast(array)
        if not isinstance(mark, MarkedJson):
            return None

        interface = mark.interface.match_by_name()
        if interface is not None and interface.check(mark.value):
            return interface
        return None

    @classmethod
    def match(cls, array: Any, fast: bool = False) -> Type["Interface"]:
        """
        Find the interface that should be used for this array based on its input type

        First runs the ``check`` method for all interfaces returned by
        :meth:`.Interface.interfaces` **except** for :class:`.NumpyInterface` ,
        and if no match is found then try the numpy interface. This is because
        :meth:`.NumpyInterface.check` can be expensive, as we could potentially
        try to

        Args:
            fast (bool): if ``False`` , check all interfaces and raise exceptions for
              having multiple matching interfaces (default). If ``True`` ,
              check each interface (as ordered by its ``priority`` , decreasing),
              and return on the first match.
        """
        # Shortcircuit match if this is a marked json dump
        array = MarkedJson.try_cast(array)
        if (match := cls.match_mark(array)) is not None:
            return match
        elif isinstance(array, MarkedJson):
            array = array.value

        # first try and find a non-numpy interface, since the numpy interface
        # will try and load the array into memory in its check method
        interfaces = cls.interfaces()
        non_np_interfaces = [i for i in interfaces if i.name != "numpy"]
        np_interface = [i for i in interfaces if i.name == "numpy"][0]

        if fast:
            matches = []
            for i in non_np_interfaces:
                if i.check(array):
                    return i
        else:
            matches = [i for i in non_np_interfaces if i.check(array)]

        if len(matches) > 1:
            msg = f"More than one interface matches input {array}:\n"
            msg += "\n".join([f"  - {i}" for i in matches])
            raise TooManyMatchesError(msg)
        elif len(matches) == 0:
            # now try the numpy interface
            if np_interface.check(array):
                return np_interface
            else:
                raise NoMatchError(f"No matching interfaces found for input {array}")
        else:
            return matches[0]

    @classmethod
    def match_output(cls, array: Any) -> Type["Interface"]:
        """
        Find the interface that should be used based on the output type -
        in the case that the output type differs from the input type, eg.
        the HDF5 interface, match an instantiated array for purposes of
        serialization to json, etc.
        """
        matches = [i for i in cls.interfaces() if isinstance(array, i.return_type)]
        if len(matches) > 1:
            msg = f"More than one interface matches output {array}:\n"
            msg += "\n".join([f"  - {i}" for i in matches])
            raise TooManyMatchesError(msg)
        elif len(matches) == 0:
            raise NoMatchError(f"No matching interfaces found for output {array}")
        else:
            return matches[0]

    @classmethod
    @lru_cache(maxsize=32)
    def mark_interface(cls) -> InterfaceMark:
        """
        Create an interface mark indicating this interface for validation after
        JSON serialization with ``round_trip==True``
        """
        interface_module = inspect.getmodule(cls)
        interface_module = (
            None if interface_module is None else interface_module.__name__
        )
        try:
            v = (
                None
                if interface_module is None
                else version(interface_module.split(".")[0])
            )
        except (
            PackageNotFoundError
        ):  # pragma: no cover - no tests for missing interface deps
            v = None

        return InterfaceMark(
            module=interface_module, cls=cls.__name__, name=cls.name, version=v
        )



"""
Interfaces for HDF5 Datasets

.. note::

    HDF5 arrays are accessed through a proxy class :class:`.H5Proxy` .
    Getting/setting values should work as normal, **except** that setting
    values on nested views is impossible -

    Specifically this doesn't work:

    .. code-block:: python

        my_model.array[0][0] = 1

    But this does work:

    .. code-block:: python

        my_model.array[0,0] = 1

    To have direct access to the hdf5 dataset, use the
    :meth:`.H5Proxy.open` method.

Datetimes
---------

Datetimes are supported as a dtype annotation, but currently they must be stored
as ``S32`` isoformatted byte strings (timezones optional) like:

.. code-block:: python

    import h5py
    from datetime import datetime
    import numpy as np
    data = np.array([datetime.now().isoformat().encode('utf-8')], dtype="S32")
    h5f = h5py.File('test.hdf5', 'w')
    h5f.create_dataset('data', data=data)

"""

import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Iterable, List, NamedTuple, Optional, Tuple, TypeVar, Union

import numpy as np
from pydantic import SerializationInfo

from numpydantic.interface.interface import Interface, JsonDict
from numpydantic.types import DtypeType, NDArrayType

try:
    import h5py
except ImportError:  # pragma: no cover
    h5py = None

if sys.version_info.minor >= 10:
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

H5Arraylike: TypeAlias = Tuple[Union[Path, str], str]

T = TypeVar("T")


class H5ArrayPath(NamedTuple):
    """Location specifier for arrays within an HDF5 file"""

    file: Union[Path, str]
    """Location of HDF5 file"""
    path: str
    """Path within the HDF5 file"""
    field: Optional[Union[str, List[str]]] = None
    """Refer to a specific field within a compound dtype"""


class H5JsonDict(JsonDict):
    """Round-trip Json-able version of an HDF5 dataset"""

    file: str
    path: str
    field: Optional[str] = None

    def to_array_input(self) -> H5ArrayPath:
        """Construct an :class:`.H5ArrayPath`"""
        return H5ArrayPath(
            **{k: v for k, v in self.model_dump().items() if k in H5ArrayPath._fields}
        )


class H5Proxy:
    """
    Proxy class to mimic numpy-like array behavior with an HDF5 array

    The attribute and item access methods only open the file for the duration of the
    method, making it less perilous to share this object between threads and processes.

    This class attempts to be a passthrough class to a :class:`h5py.Dataset` object,
    including its attributes and item getters/setters.

    When using read-only methods, no locking is attempted (beyond the HDF5 defaults),
    but when using the write methods (setting an array value), try and use the
    ``locking`` methods of :class:`h5py.File` .

    Args:
        file (pathlib.Path | str): Location of hdf5 file on filesystem
        path (str): Path to array within hdf5 file
        field (str, list[str]): Optional - refer to a specific field within
            a compound dtype
        annotation_dtype (dtype): Optional - the dtype of our type annotation
    """

    def __init__(
        self,
        file: Union[Path, str],
        path: str,
        field: Optional[Union[str, List[str]]] = None,
        annotation_dtype: Optional[DtypeType] = None,
    ):
        self._h5f = None
        self.file = Path(file).resolve()
        self.path = path
        self.field = field
        self._annotation_dtype = annotation_dtype
        self._h5arraypath = H5ArrayPath(self.file, self.path, self.field)

    def array_exists(self) -> bool:
        """Check that there is in fact an array at :attr:`.path` within :attr:`.file`"""
        with h5py.File(self.file, "r") as h5f:
            obj = h5f.get(self.path)
            return obj is not None

    @classmethod
    def from_h5array(cls, h5array: H5ArrayPath) -> "H5Proxy":
        """Instantiate using :class:`.H5ArrayPath`"""
        return H5Proxy(file=h5array.file, path=h5array.path, field=h5array.field)

    @property
    def dtype(self) -> np.dtype:
        """
        Get dtype of array, using :attr:`.field` if present
        """
        with h5py.File(self.file, "r") as h5f:
            obj = h5f.get(self.path)
            if self.field is None:
                return obj.dtype
            else:
                return obj.dtype[self.field]

    def __array__(self) -> np.ndarray:
        """To a numpy array"""
        with h5py.File(self.file, "r") as h5f:
            obj = h5f.get(self.path)
            return obj[:]

    def __getattr__(self, item: str):
        if item == "__name__":
            # special case for H5Proxies that don't refer to a real file during testing
            return "H5Proxy"
        with h5py.File(self.file, "r") as h5f:
            obj = h5f.get(self.path)
            val = getattr(obj, item)
            return val

    def __getitem__(
        self, item: Union[int, slice, Tuple[Union[int, slice], ...]]
    ) -> Union[np.ndarray, DtypeType]:
        with h5py.File(self.file, "r") as h5f:
            obj = h5f.get(self.path)
            # handle compound dtypes
            if self.field is not None:
                # handle compound string dtype
                if encoding := h5py.h5t.check_string_dtype(obj.dtype[self.field]):
                    if isinstance(item, tuple):
                        item = (*item, self.field)
                    else:
                        item = (item, self.field)

                    try:
                        # single string
                        val = obj[item].decode(encoding.encoding)
                        if self._annotation_dtype is np.datetime64:
                            return np.datetime64(val)
                        else:
                            return val
                    except AttributeError:
                        # numpy array of bytes
                        val = np.char.decode(obj[item], encoding=encoding.encoding)
                        if self._annotation_dtype is np.datetime64:
                            return val.astype(np.datetime64)
                        else:
                            return val
                # normal compound type
                else:
                    obj = obj.fields(self.field)
            else:
                if h5py.h5t.check_string_dtype(obj.dtype):
                    obj = obj.asstr()

            val = obj[item]
            if self._annotation_dtype is np.datetime64:
                if isinstance(val, str):
                    return np.datetime64(val)
                else:
                    return val.astype(np.datetime64)
            else:
                return val

    def __setitem__(
        self,
        key: Union[int, slice, Tuple[Union[int, slice], ...]],
        value: Union[int, float, datetime, np.ndarray],
    ):
        # TODO: Make a generalized value serdes system instead of ad-hoc type conversion
        value = self._serialize_datetime(value)
        with h5py.File(self.file, "r+", locking=True) as h5f:
            obj = h5f.get(self.path)
            if self.field is None:
                obj[key] = value
            else:
                if isinstance(key, tuple):
                    key = (*key, self.field)
                    obj[key] = value
                else:
                    obj[key, self.field] = value

    def __len__(self) -> int:
        """self.shape[0]"""
        return self.shape[0]

    def __eq__(self, other: "H5Proxy") -> bool:
        """
        Check that we are referring to the same hdf5 array
        """
        if isinstance(other, H5Proxy):
            return self._h5arraypath == other._h5arraypath
        else:
            raise ValueError("Can only compare equality of two H5Proxies")

    def open(self, mode: str = "r") -> "h5py.Dataset":
        """
        Return the opened :class:`h5py.Dataset` object

        You must remember to close the associated file with :meth:`.close`
        """
        if self._h5f is None:
            self._h5f = h5py.File(self.file, mode)
        return self._h5f.get(self.path)

    def close(self) -> None:
        """
        Close the :class:`h5py.File` object left open when returning the dataset with
        :meth:`.open`
        """
        if self._h5f is not None:
            self._h5f.close()
        self._h5f = None

    def _serialize_datetime(self, v: Union[T, datetime]) -> Union[T, bytes]:
        """
        Convert a datetime into a bytestring
        """
        if self._annotation_dtype is np.datetime64:
            if not isinstance(v, Iterable):
                v = [v]
            v = np.array(v).astype("S32")
        return v


class H5Interface(Interface):
    """
    Interface for Arrays stored as datasets within an HDF5 file.

    Takes a :class:`.H5ArrayPath` specifier to select a :class:`h5py.Dataset` from a
    :class:`h5py.File` and returns a :class:`.H5Proxy` class that acts like a
    passthrough numpy-like interface to the dataset.
    """

    name = "hdf5"
    input_types = (H5ArrayPath, H5Arraylike, H5Proxy)
    return_type = H5Proxy
    json_model = H5JsonDict

    @classmethod
    def enabled(cls) -> bool:
        """Check whether h5py can be imported"""
        return h5py is not None

    @classmethod
    def check(cls, array: Union[H5ArrayPath, Tuple[Union[Path, str], str]]) -> bool:
        """
        Check that the given array is a :class:`.H5ArrayPath` or something that
        resembles one.
        """
        if isinstance(array, (H5ArrayPath, H5Proxy)):
            return True

        if isinstance(array, dict):
            if array.get("type", False) == cls.name:
                return True
            # continue checking if dict contains an hdf5 file
            file = array.get("file", "")
            array = (file, "")

        if isinstance(array, (tuple, list)) and len(array) in (2, 3):
            # check that the first arg is an hdf5 file
            try:
                file = Path(array[0])
            except TypeError:
                # not a path, we don't apply.
                return False

            if not file.exists():
                return False

            # hdf5 files are commonly given odd suffixes,
            # so we just try and open it and see what happens
            try:
                with h5py.File(file, "r"):
                    # don't check that the array exists and raise here,
                    # this check is just for whether the validator applies or not.
                    pass
                return True
            except (FileNotFoundError, OSError):
                return False

        return False

    def before_validation(self, array: Any) -> NDArrayType:
        """Create an :class:`.H5Proxy` to use throughout validation"""
        if isinstance(array, H5ArrayPath):
            array = H5Proxy.from_h5array(h5array=array)
        elif isinstance(array, H5Proxy):
            # nothing to do, already proxied
            pass
        elif isinstance(array, (tuple, list)) and len(array) == 2:  # pragma: no cover
            array = H5Proxy(file=array[0], path=array[1])
        elif isinstance(array, (tuple, list)) and len(array) == 3:
            array = H5Proxy(file=array[0], path=array[1], field=array[2])
        else:  # pragma: no cover
            # this should never happen really since `check` confirms this before
            # we'd reach here, but just to complete the if else...
            raise ValueError(
                "Need to specify a file and a path within an HDF5 file to use the HDF5 "
                "Interface"
            )
        array._annotation_dtype = self.dtype

        if not array.array_exists():
            raise ValueError(
                f"HDF5 file located at {array.file}, "
                f"but no array found at {array.path}"
            )

        return array

    def get_dtype(self, array: NDArrayType) -> DtypeType:
        """
        Get the dtype from the input array

        Subclasses to correctly handle
        """
        if h5py.h5t.check_string_dtype(array.dtype):
            # check for datetimes
            try:
                if array[0].dtype.type is np.datetime64:
                    return np.datetime64
                else:
                    return str
            except (AttributeError, TypeError):  # pragma: no cover
                # it's not a datetime, but it is some kind of string
                return str
            except (IndexError, ValueError):
                # if the dataset is empty, we can't tell if something is a datetime
                # or not, so we just tell the validation method what it wants to hear
                if self.dtype in (np.datetime64, str):
                    return self.dtype
                else:
                    return str
        else:
            return array.dtype

    @classmethod
    def to_json(cls, array: H5Proxy, info: Optional[SerializationInfo] = None) -> dict:
        """
        Render HDF5 array as JSON

        If ``round_trip == True``, we dump just the proxy info, a dictionary like:

        * ``file``: :attr:`.file`
        * ``path``: :attr:`.path`
        * ``attrs``: Any HDF5 attributes on the dataset
        * ``array``: The array as a list of lists

        Otherwise, we dump the array as a list of lists
        """
        if info.round_trip:
            as_json = {
                "type": cls.name,
            }
            as_json.update(array._h5arraypath._asdict())
        else:
            try:
                dset = array.open()
                as_json = dset[:].tolist()
            finally:
                array.close()

        return as_json



"""
Tests for the testing helpers lmao
"""

import numpy as np
import pytest
from pydantic import BaseModel

from numpydantic import NDArray, Shape
from numpydantic.testing.cases import INTERFACE_CASES
from numpydantic.testing.helpers import ValidationCase
from numpydantic.testing.interfaces import NumpyCase


def test_validation_case_merge():
    case_1 = ValidationCase(id="1", interface=NumpyCase, passes=False)
    case_2 = ValidationCase(id="2", dtype=str, passes=True)
    case_3 = ValidationCase(id="3", shape=(1, 2, 3), passes=True)

    merged_simple = case_2.merge(case_3)
    assert merged_simple.dtype == case_2.dtype
    assert merged_simple.shape == case_3.shape

    merged_multi = case_1.merge([case_2, case_3])
    assert merged_multi.dtype == case_2.dtype
    assert merged_multi.shape == case_3.shape
    assert merged_multi.interface == case_1.interface

    # passes should be true only if all the cases are
    assert merged_simple.passes
    assert not merged_multi.passes

    # ids should merge
    assert merged_simple.id == "2-3"
    assert merged_multi.id == "1-2-3"


@pytest.mark.parametrize(
    "interface",
    [
        pytest.param(
            i.interface, marks=getattr(pytest.mark, i.interface.interface.name)
        )
        for i in INTERFACE_CASES
        if i.id not in ("hdf5_compound")
    ],
)
def test_make_array(interface, tmp_output_dir_func):
    """
    An interface case can generate an array from params or a given array

    Not testing correctness here, that's what hte rest of the testing does.
    """
    arr = np.zeros((10, 10, 2, 3), dtype=np.uint8)
    arr = interface.make_array(array=arr, dtype=np.uint8, path=tmp_output_dir_func)

    class MyModel(BaseModel):
        array: NDArray[Shape["10, 10, 2, 3"], np.uint8]

    _ = MyModel(array=arr)



"""
Serialization helpers for :func:`pydantic.BaseModel.model_dump`
and :func:`pydantic.BaseModel.model_dump_json` .
"""

from pathlib import Path
from typing import Any, Callable, Iterable, TypeVar, Union

from pydantic_core.core_schema import SerializationInfo

from numpydantic.interface import Interface, JsonDict

T = TypeVar("T")
U = TypeVar("U")


def jsonize_array(value: Any, info: SerializationInfo) -> Union[list, dict]:
    """Use an interface class to render an array as JSON"""
    # perf: keys to skip in generation - anything named "value" is array data.
    skip = ["value"]

    interface_cls = Interface.match_output(value)
    array = interface_cls.to_json(value, info)
    if isinstance(array, JsonDict):
        array = array.model_dump(exclude_none=True)

    if info.context:
        if info.context.get("mark_interface", False):
            array = interface_cls.mark_json(array)

        if isinstance(array, list):
            return array

        # ---- Perf Barrier ------------------------------------------------------
        # put context args intended to **wrap** the array above
        # put context args intended to **modify** the array below
        #
        # above, we assume that a list is **data** not to be modified.
        # below, we must mark whenever the data is in the line of fire
        # to avoid an expensive iteration.

        if info.context.get("absolute_paths", False):
            array = _absolutize_paths(array, skip)
        else:
            relative_to = info.context.get("relative_to", ".")
            array = _relativize_paths(array, relative_to, skip)
    else:
        if isinstance(array, list):
            return array

        # ---- Perf Barrier ------------------------------------------------------
        # same as above, ensure any keys that contain array values are skipped right now

        array = _relativize_paths(array, ".", skip)

    return array


def _relativize_paths(
    value: dict, relative_to: str = ".", skip: Iterable = tuple()
) -> dict:
    """
    Make paths relative to either the current directory or the provided
    ``relative_to`` directory, if provided in the context
    """
    relative_to = Path(relative_to).resolve()

    def _r_path(v: Any) -> Any:
        if not isinstance(v, (str, Path)):
            return v
        try:
            path = Path(v)
            resolved = path.resolve()
            # skip things that are pathlike but either don't exist
            # or that are at the filesystem root (eg like /data)
            if (
                not path.exists()
                or (resolved.is_dir() and str(resolved.parent) == resolved.anchor)
                or relative_to.anchor != resolved.anchor
            ):
                return v
            return str(relative_path(path, relative_to))
        except (TypeError, ValueError, OSError):
            return v

    return _walk_and_apply(value, _r_path, skip)


def _absolutize_paths(value: dict, skip: Iterable = tuple()) -> dict:
    def _a_path(v: Any) -> Any:
        if not isinstance(v, (str, Path)):
            return v
        try:
            path = Path(v)
            if not path.exists():
                return v
            return str(path.resolve())
        except (TypeError, ValueError, OSError):
            return v

    return _walk_and_apply(value, _a_path, skip)


def _walk_and_apply(value: T, f: Callable[[U, bool], U], skip: Iterable = tuple()) -> T:
    """
    Walk an object, applying a function
    """
    if isinstance(value, dict):
        for k, v in value.items():
            if k in skip:
                continue
            if isinstance(v, dict):
                _walk_and_apply(v, f, skip)
            elif isinstance(v, list):
                value[k] = [_walk_and_apply(sub_v, f, skip) for sub_v in v]
            else:
                value[k] = f(v)
    elif isinstance(value, list):
        value = [_walk_and_apply(v, f, skip) for v in value]
    else:
        value = f(value)
    return value


def relative_path(self: Path, other: Path, walk_up: bool = True) -> Path:
    """
    "Backport" of :meth:`pathlib.Path.relative_to` with ``walk_up=True``
    that's not available pre 3.12.

    Return the relative path to another path identified by the passed
    arguments.  If the operation is not possible (because this is not
    related to the other path), raise ValueError.

    The *walk_up* parameter controls whether `..` may be used to resolve
    the path.

    References:
        https://github.com/python/cpython/blob/8a2baedc4bcb606da937e4e066b4b3a18961cace/Lib/pathlib/_abc.py#L244-L270
    """
    if not isinstance(other, Path):  # pragma: no cover - ripped from cpython
        other = Path(other)
    self_parts = self.parts
    other_parts = other.parts
    anchor0, parts0 = self_parts[0], list(reversed(self_parts[1:]))
    anchor1, parts1 = other_parts[0], list(reversed(other_parts[1:]))
    if anchor0 != anchor1:
        raise ValueError(f"{self!r} and {other!r} have different anchors")
    while parts0 and parts1 and parts0[-1] == parts1[-1]:
        parts0.pop()
        parts1.pop()
    for part in parts1:  # pragma: no cover - not testing, ripped off from cpython
        if not part or part == ".":
            pass
        elif not walk_up:
            raise ValueError(f"{self!r} is not in the subpath of {other!r}")
        elif part == "..":
            raise ValueError(f"'..' segment in {other!r} cannot be walked")
        else:
            parts0.append("..")
    return Path(*reversed(parts0))
